{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FunctionalDependencies #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE UndecidableInstances #-}
-----------------------------------------------------------------------------
-- |
-- Module: Data.Enumerator.Instances.FD
-- Copyright: 2010 John Millikin
-- License: MIT
--
-- Maintainer: jmillikin@gmail.com
-- Portability: portable
--
-- Enumerator instances for monads-fd classes
--
-----------------------------------------------------------------------------
module Data.Enumerator.Instances.FD () where

import qualified Control.Exception as Exc

import Control.Monad.Trans (lift)
import qualified Control.Monad.Cont.Class as M_C
import qualified Control.Monad.Error.Class as M_E
import qualified Control.Monad.RWS.Class as M_RWS
import qualified Control.Monad.Reader.Class as M_R
import qualified Control.Monad.State.Class as M_S
import qualified Control.Monad.Writer.Class as M_W

import qualified Data.Enumerator as E
import Data.Monoid (mempty, mappend)

instance M_C.MonadCont m => M_C.MonadCont (E.Iteratee a m) where
	callCC f = E.Iteratee $ M_C.callCC $ \c -> let
		emptyYield x = E.Yield x (E.Chunks [])
		in E.runIteratee (f (E.Iteratee . c . emptyYield))

instance Monad m => M_E.MonadError Exc.SomeException (E.Iteratee a m) where
	throwError = E.throwError
	catchError = E.catchError

instance M_RWS.MonadRWS r w s m => M_RWS.MonadRWS r w s (E.Iteratee a m)

instance M_R.MonadReader r m => M_R.MonadReader r (E.Iteratee a m) where
	ask = lift M_R.ask
	local f m = E.Iteratee (M_R.local f (E.runIteratee m))

instance M_S.MonadState s m =>  M_S.MonadState s (E.Iteratee a m) where
	get = lift  M_S.get
	put = lift .  M_S.put

instance M_W.MonadWriter w m => M_W.MonadWriter w (E.Iteratee a m) where
	tell = lift . M_W.tell
	
	listen = loop mempty where
		loop w0 m = E.Iteratee $ do
			~(step, w1) <- M_W.listen (E.runIteratee m)
			let w = mappend w0 w1
			return $ case step of
				E.Yield x cs -> E.Yield (x, w) cs
				E.Error err -> E.Error err
				E.Continue k -> E.Continue (loop w . k)
	
	pass m = E.Iteratee $ M_W.pass $ do
		step <- E.runIteratee m
		case step of
			E.Yield (x, wf) cs -> return (E.Yield x cs, wf)
			E.Error err -> return (E.Error err, id)
			E.Continue k -> return (E.Continue (M_W.pass . k), id)
